from http import HTTPStatus

from django.conf import settings

DEFAULT_LOGBOX_SETTINGS = {
    # HTTP methods to log. Default to all
    "LOGGING_HTTP_METHODS": ["GET", "POST", "PUT", "PATCH", "DELETE"],
    # exclude server IPs from logging. Default to not exclude any
    "LOGGING_SERVER_IPS_TO_EXCLUDE": [],
    # exclude client IPs from logging. Default to not exclude any
    "LOGGING_CLIENT_IPS_TO_EXCLUDE": [],
    # Status codes to log. Default to all
    "LOGGING_STATUS_CODES": [http_code.value for http_code in HTTPStatus],
    # Path regex to exclude from logging. Default to not exclude any
    "LOGGING_PATHS_TO_EXCLUDE": [],
    # The number of logs to insert in bulk. The default is 1, which means insert logs instantly.
    "LOGGING_DAEMON_QUEUE_SIZE": 1,
    # The number of seconds between log insertion attempts. The default is 0.
    "LOGGING_DAEMON_INTERVAL": 0,
}

USER_LOGBOX_SETTINGS = getattr(settings, "LOGBOX_SETTINGS", {})
settings.LOGBOX_SETTINGS = {**DEFAULT_LOGBOX_SETTINGS, **USER_LOGBOX_SETTINGS}



from django.http import HttpRequest

from django_logbox.logging import add_log


class LogboxMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request: HttpRequest):
        response = self.get_response(request)

        add_log(
            request=request,
            response=response,
            exception=None,
        )

        return response

    def process_exception(self, request: HttpRequest, exception: Exception):
        add_log(
            request=request,
            response=None,
            exception=exception,
        )



import traceback

from django.http import HttpRequest, HttpResponse
from ua_parser import Result

########################
# Django request utils #
########################


def get_method(request: HttpRequest):
    return request.method


def get_path(request: HttpRequest):
    return request.path


def get_status_code(response: HttpResponse):
    return response.status_code


def get_user_agent(request: HttpRequest) -> str | None:
    return request.META.get("HTTP_USER_AGENT", None)


def get_querystring(request: HttpRequest):
    return (
        None
        if request.META.get("QUERY_STRING", None) == ""
        else request.META.get("QUERY_STRING", None)
    )


def get_request_body(request: HttpRequest):
    return request.body.decode("utf-8") if request.body else None


def get_exception_type(exception: Exception) -> str:
    return exception.__class__.__name__


def get_traceback(exception: Exception) -> str:
    return "".join(traceback.format_tb(exception.__traceback__))


def get_server_ip(request: HttpRequest):
    return request.get_host()


def get_client_ip(request: HttpRequest):
    x_forwarded_for = request.META.get("HTTP_X_FORWARDED_FOR")
    if x_forwarded_for:
        ip = x_forwarded_for.split(",")[0]
    else:
        ip = request.META.get("REMOTE_ADDR")
    return ip


############################
# User-Agent parsing utils #
#############################


def device_str(parsed_useragent: Result) -> str | None:
    """
    extract device data from parsed useragent.

    for example: Mac(Apple, Mac), Samsung SM-S711N(Samsung, SM-S711N)
    """
    if not parsed_useragent.device:
        return None

    family = parsed_useragent.device.family if parsed_useragent.device.family else ""

    brand = parsed_useragent.device.brand if parsed_useragent.device.brand else ""
    model = parsed_useragent.device.model if parsed_useragent.device.model else ""

    return f"{family}({brand}, {model})" if brand and model else family


def os_str(parsed_useragent: Result) -> str | None:
    """
    extract OS data from parsed useragent.

    for example: Windows(10), Mac OS X(10.15.7), SomeOS
    """
    if not parsed_useragent.os:
        return None

    family = parsed_useragent.os.family if parsed_useragent.os.family else ""

    version_information = [
        parsed_useragent.os.major,
        parsed_useragent.os.minor,
        parsed_useragent.os.patch,
        parsed_useragent.os.patch_minor,
    ]
    version_suffix = ".".join(list(filter(None, version_information)))

    return f"{family}({version_suffix})" if version_suffix else family


def browser_str(parsed_useragent: Result) -> str | None:
    """
    extract browser data from parsed useragent.

    for example: Chrome(86), Safari(13.1.2), SomeBrowser
    """
    if not parsed_useragent.user_agent:
        return None

    family = (
        parsed_useragent.user_agent.family if parsed_useragent.user_agent.family else ""
    )

    version_information = [
        parsed_useragent.user_agent.major,
        parsed_useragent.user_agent.minor,
        parsed_useragent.user_agent.patch,
        parsed_useragent.user_agent.patch_minor,
    ]
    version_suffix = ".".join(list(filter(None, version_information)))

    return f"{family}({version_suffix})" if version_suffix else family



# Generated by Django 4.2.19 on 2025-03-22 16:34

from django.db import migrations
from ua_parser import parse

from django_logbox.utils import browser_str, device_str, os_str


def _populate_device_os_browser_fields(apps, schema_editor):
    ServerLog = apps.get_model("django_logbox", "ServerLog")
    for serverlog in ServerLog.objects.all():
        parsed_useragent_result = parse(serverlog.user_agent)

        serverlog.device = device_str(parsed_useragent_result)
        serverlog.os = os_str(parsed_useragent_result)
        serverlog.browser = browser_str(parsed_useragent_result)

        serverlog.save()


class Migration(migrations.Migration):
    dependencies = [
        ("django_logbox", "0005_alter_serverlog_options_serverlog_browser_and_more"),
    ]

    operations = [
        migrations.RunPython(_populate_device_os_browser_fields),
    ]



# Generated by Django 4.2.19 on 2025-03-31 12:17

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("auth", "__first__"),
        ("django_logbox", "0006_backfill_device_os_browser_fields"),
    ]

    operations = [
        migrations.AddField(
            model_name="serverlog",
            name="user",
            field=models.ForeignKey(
                blank=True,
                help_text="User associated with the request, if authenticated.",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="server_logs",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AlterField(
            model_name="serverlog",
            name="querystring",
            field=models.TextField(
                help_text="Query parameters of the request as a URL-encoded string, e.g., 'param1=value1&ampparam2=value2'.",
                null=True,
                verbose_name="querystring",
            ),
        ),
    ]



import re

from django.http import HttpRequest, HttpResponse

from django_logbox.app_settings import settings
from django_logbox.utils import get_client_ip, get_method, get_server_ip


class LogboxLogFilter:
    @staticmethod
    def should_filter_log(
        request: HttpRequest,
        response: HttpResponse | None,
    ) -> bool:
        """
        Check if the request and response should be logged.

        :return: True if the request and response should be filtered, False otherwise.
        """
        # Check if the request is already logged
        if hasattr(request, "logbox_logged"):
            return True

        # Check if the request and response should be logged
        if not LogboxLogFilter.should_log_request(request=request):
            return True
        if response:
            if not LogboxLogFilter.should_log_response(response=response):
                return True

        return False

    @staticmethod
    def should_log_request(request: HttpRequest) -> bool:
        return (
            LogboxLogFilter.is_client_ip_allowed(
                client_ip=get_client_ip(request=request)
            )
            and LogboxLogFilter.is_server_ip_allowed(
                server_ip=get_server_ip(request=request)
            )
            and LogboxLogFilter.is_method_allowed(
                method=get_method(request=request),
            )
            and LogboxLogFilter.is_path_allowed(
                request_path=request.path,
            )
        )

    @staticmethod
    def is_method_allowed(method: str) -> bool:
        """
        Check if the HTTP method is allowed to be logged.
        """
        return method in settings.LOGBOX_SETTINGS["LOGGING_HTTP_METHODS"]

    @staticmethod
    def is_client_ip_allowed(client_ip: str) -> bool:
        """
        Check if the client IP is allowed to be logged.
        """
        return (
            client_ip not in settings.LOGBOX_SETTINGS["LOGGING_CLIENT_IPS_TO_EXCLUDE"]
        )

    @staticmethod
    def is_server_ip_allowed(server_ip: str) -> bool:
        """
        Check if the server IP is allowed to be logged.
        """
        return (
            server_ip not in settings.LOGBOX_SETTINGS["LOGGING_SERVER_IPS_TO_EXCLUDE"]
        )

    @staticmethod
    def is_path_allowed(request_path: str) -> bool:
        """Filter requests based on path patterns."""

        return not any(
            re.match(path, request_path)
            for path in settings.LOGBOX_SETTINGS["LOGGING_PATHS_TO_EXCLUDE"]
        )

    @staticmethod
    def should_log_response(response: HttpResponse):
        return response.status_code in settings.LOGBOX_SETTINGS["LOGGING_STATUS_CODES"]



# Generated by Django 4.2.19 on 2025-03-22 16:33

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_logbox", "0004_alter_serverlog_client_ip_and_more"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="serverlog",
            options={
                "ordering": ("-timestamp",),
                "verbose_name": "Server Log",
                "verbose_name_plural": "Server Logs",
            },
        ),
        migrations.AddField(
            model_name="serverlog",
            name="browser",
            field=models.CharField(
                blank=True,
                help_text="Browser data parsed from User-Agent request header.",
                max_length=255,
                null=True,
                verbose_name="browser",
            ),
        ),
        migrations.AddField(
            model_name="serverlog",
            name="device",
            field=models.CharField(
                blank=True,
                help_text="Device data parsed from User-Agent request header.",
                max_length=255,
                null=True,
                verbose_name="device",
            ),
        ),
        migrations.AddField(
            model_name="serverlog",
            name="os",
            field=models.CharField(
                blank=True,
                help_text="OS data parsed from User-Agent request header.",
                max_length=255,
                null=True,
                verbose_name="os",
            ),
        ),
        migrations.AlterField(
            model_name="serverlog",
            name="user_agent",
            field=models.TextField(
                help_text="User-Agent string from the client's request header, providing browser and OS details.",
                max_length=255,
                null=True,
                verbose_name="user_agent",
            ),
        ),
    ]



from django.contrib import admin
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

from .models import ServerLog


@admin.register(ServerLog)
class ServerLogAdmin(admin.ModelAdmin):
    list_display = (
        "method",
        "path",
        "status_code",
        "device",
        "timestamp",
        "server_ip",
        "client_ip",
        "user",
    )

    readonly_fields = (
        "method",
        "path",
        "status_code",
        "user_agent",
        "querystring",
        "request_body",
        "timestamp",
        "exception_type",
        "exception_message",
        "traceback",
        "server_ip",
        "client_ip",
    )

    fieldsets = (
        (
            _("Request Information"),
            {
                "fields": (
                    "timestamp",
                    "method",
                    "path",
                    "user_agent_details",
                    "status_code",
                    "querystring",
                    "request_body",
                    "user",
                ),
            },
        ),
        (
            _("Exception Details"),
            {
                "fields": (
                    "exception_type",
                    "exception_message",
                    "traceback",
                ),
            },
        ),
        (
            _("IP Addresses"),
            {
                "fields": (
                    "server_ip",
                    "client_ip",
                ),
            },
        ),
    )
    list_display_links = (
        "method",
        "path",
    )
    search_fields = (
        "status_code",
        "exception_message",
        "client_ip",
        "server_ip",
    )
    list_filter = (
        "method",
        "status_code",
        "path",
        "timestamp",
    )
    raw_id_fields = ("user",)

    change_list_template = "change_list.html"

    def changelist_view(self, request, extra_context: dict = None):
        extra_context = {
            "traffic_data": {
                "label_data": [
                    data["date"].isoformat()
                    for data in ServerLog.objects.get_traffic_data()
                ],
                "count_data": [
                    data["count"] for data in ServerLog.objects.get_traffic_data()
                ],
            },
            "device_data": {
                entry["device"]: {
                    "count": entry["count"],
                    "percentage": entry["percentage"],
                }
                for entry in ServerLog.objects.get_device_data()
            },
            "os_data": {
                entry["os"]: {
                    "count": entry["count"],
                    "percentage": entry["percentage"],
                }
                for entry in ServerLog.objects.get_os_data()
            },
            "browser_data": {
                entry["browser"]: {
                    "count": entry["count"],
                    "percentage": entry["percentage"],
                }
                for entry in ServerLog.objects.get_browser_data()
            },
        }
        return super().changelist_view(
            request,
            extra_context=extra_context,
        )

    @admin.display(description=_("User-Agent details"))
    def user_agent_details(self, obj):
        return format_html(
            f"<p>{obj.user_agent}</p>"
            f"<li>Device: {obj.device}</li>"
            f"<li>OS: {obj.os}</li>"
            f"<li>Browser: {obj.browser}</li>"
        )

    def get_queryset(self, request):
        queryset = super().get_queryset(request)
        return queryset.select_related("user")

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False



import logging
import signal
import threading
import time
from queue import Queue
from types import FrameType

from django_logbox.app_settings import settings

logger = logging.getLogger("logbox")


class ServerLogInsertThread(threading.Thread):
    def __init__(
        self,
        logging_daemon_interval: int,
        logging_daemon_queue_size: int,
    ):
        super().__init__(
            name="logbox_logger_thread",
            daemon=True,
        )
        from django_logbox.models import ServerLog

        self._serverlog_model = ServerLog
        self._logging_daemon_interval = logging_daemon_interval
        self._logging_daemon_queue_size = logging_daemon_queue_size
        self._queue = Queue(maxsize=self._logging_daemon_queue_size)

        self._stop_event = threading.Event()
        signal.signal(signal.SIGINT, self._exit_gracefully)
        signal.signal(signal.SIGTERM, self._exit_gracefully)

    def run(self) -> None:
        """
        Continuously runs the logging thread, periodically inserting logs in bulk.

        This method sleeps for the specified interval (`_logging_daemon_interval`)
        and then triggers the bulk insertion of logs from the queue. If an exception
        occurs during the process, it logs the error.
        """
        while not self._stop_event.is_set():
            try:
                time.sleep(self._logging_daemon_interval)
                self._start_bulk_insertion()
            except Exception as e:
                logger.error(f"Error occurred while inserting logs: {e}")

    def put_serverlog(self, data) -> None:
        self._queue.put(self._serverlog_model(**data))
        if self._queue.qsize() >= self._logging_daemon_queue_size:
            logger.debug(
                f"Queue is full({self._queue.qsize()}), starting bulk insertion"
            )
            self._start_bulk_insertion()

    def _start_bulk_insertion(self):
        bulk_item = []
        while not self._queue.empty():
            bulk_item.append(self._queue.get())
        if bulk_item:
            self._serverlog_model.objects.bulk_create(bulk_item)

    def _exit_gracefully(
        self,
        sig: int,
        frame: FrameType | None,
    ) -> None:
        logger.info(f"Received signal {sig}. Exiting gracefully...")
        self._stop_event.set()
        self._start_bulk_insertion()
        logger.info("All logs have been inserted. Exiting.")
        exit(0)


logbox_logger_thread = ServerLogInsertThread(
    logging_daemon_interval=settings.LOGBOX_SETTINGS["LOGGING_DAEMON_INTERVAL"],
    logging_daemon_queue_size=settings.LOGBOX_SETTINGS["LOGGING_DAEMON_QUEUE_SIZE"],
)



# Generated by Django 5.1.1 on 2024-10-07 14:02

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("django_logbox", "0001_initial"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="serverlog",
            name="duration",
        ),
    ]



from datetime import datetime
from http import HTTPStatus

from django.http import HttpRequest, HttpResponse
from django.utils import timezone
from ua_parser import parse

from django_logbox.filtering import LogboxLogFilter
from django_logbox.threading import logbox_logger_thread
from django_logbox.utils import (
    browser_str,
    device_str,
    get_client_ip,
    get_exception_type,
    get_method,
    get_path,
    get_querystring,
    get_request_body,
    get_server_ip,
    get_status_code,
    get_traceback,
    get_user_agent,
    os_str,
)


def add_log(
    request: HttpRequest,
    response: HttpResponse | None,
    exception: Exception | None = None,
) -> None:
    """
    Add log data to the serverlog.
    add queue to the server log thread and mark the request as logged.
    """
    if LogboxLogFilter.should_filter_log(request, response):
        return

    logbox_logger_thread.put_serverlog(
        data=get_log_data(
            timestamp=timezone.now(),
            request=request,
            response=response,
            exception=exception,
        )
    )
    request.logbox_logged = True


def get_log_data(
    timestamp: datetime,
    request: HttpRequest,
    response: HttpResponse | None,
    exception: Exception | None = None,
) -> dict:
    _raw_user_agent = get_user_agent(request)

    data = {
        "method": get_method(request),
        "path": get_path(request),
        "user_agent": _raw_user_agent,
        "device": device_str(parse(_raw_user_agent)) if _raw_user_agent else None,
        "os": os_str(parse(_raw_user_agent)) if _raw_user_agent else None,
        "browser": browser_str(parse(_raw_user_agent)) if _raw_user_agent else None,
        "querystring": get_querystring(request),
        "request_body": get_request_body(request),
        "timestamp": timestamp,
        "server_ip": get_server_ip(request),
        "client_ip": get_client_ip(request),
        "status_code": (
            # if response is None, return status code 500.
            # This happens when an exception occurs and response is not set from middleware.process_exception
            get_status_code(response) if response else HTTPStatus.INTERNAL_SERVER_ERROR
        ),
        "user": request.user if request.user.is_authenticated else None,
    }

    if exception:
        exception_data = {
            "exception_type": get_exception_type(exception),
            "exception_message": str(exception),
            "traceback": get_traceback(exception),
        }
        data.update(exception_data)

    return data



import os

from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class DjangoLogboxConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "django_logbox"
    verbose_name = _("Logbox")

    def ready(self):
        if self._is_main_process():
            from django_logbox.threading import logbox_logger_thread

            logbox_logger_thread.start()

    @staticmethod
    def _is_main_process() -> bool:
        """
        Check if the current process is the main process.

        This is used to ensure that the logging thread is started only in the main process.
        """
        return os.environ.get("RUN_MAIN") == "true"
